#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include "PageDumpFrame.hh"



void
safe_fread (void* buffer,
	    size_t itemSize,
	    size_t numberOfItems,
	    FILE* inputStream) {

  size_t readResult = fread(buffer,
			    itemSize,
			    numberOfItems,
			    inputStream);

  if (readResult < numberOfItems) {
    if (readResult == -1) {
      char errorString[256];
      sprintf(errorString,
	      "Died reading %d items of size %d",
	      numberOfItems,
	      itemSize);
      perror(errorString);
    } else {
      fprintf(stderr,
	      "Read only %d of %d items of size %d",
	      readResult,
	      numberOfItems,
	      itemSize);
    }
  }
}



void
safe_fwrite (void* buffer,
	     size_t itemSize,
	     size_t numberOfItems,
	     FILE* outputStream) {

  size_t writeResult = fwrite(buffer,
			      itemSize,
			      numberOfItems,
			      outputStream);
  
  if (writeResult < numberOfItems) {
    if (writeResult == -1) {
      char errorString[256];
      sprintf(errorString,
	      "Died writing %d items of size %d",
	      numberOfItems,
	      itemSize);
      perror(errorString);
    } else {
      fprintf(stderr,
	      "Wrote only %d of %d items of size %d",
	      writeResult,
	      numberOfItems,
	      itemSize);
    }
  }
}



// Initialize the frame into a state that indicates that it's not
// currently holding a page.
PageDumpFrame::PageDumpFrame () {
  pageNumber = 0;
  dirty = false;
}



// Read a record and store it in this frame.
void
PageDumpFrame::load (FILE* inputStream) {

  // Read in whether or not this record is for a dirtied page
  char dirtyField = 0;
  fscanf(inputStream, "%1c %4x", &dirtyField, &pageNumber);

  // Skip the end-of-line character.
  char endOfLineChar = fgetc(inputStream);
  switch (dirtyField) {
  case 'R':
    dirty = false;
    break;
  case 'W':
    dirty = true;
    break;
  default:
    fprintf(stderr,
	    "Invalid dirty field %c\n",
	    dirtyField);
    exit(-1);
  }
  if (pageNumber <= 0) {
    fprintf(stderr,
	    "Invalid page number %x\n",
	    pageNumber);
    exit(-1);
  }
  
  // Read the fault dump image.
  safe_fread((void*)fetchedImage,
	     sizeof(word),
	     WORDS_PER_PAGE,
	     inputStream);

  // If this record is for a dirtied page, read the evict dump image.
  if (dirty) {
    safe_fread((void*)evictedImage,
	       sizeof(word),
	       WORDS_PER_PAGE,
	       inputStream);
  }
}



// Given another frame, determine whether or not it represents the
// same page that this one does.
bool
PageDumpFrame::match (PageDumpFrame& matchFrame) {

  return (pageNumber == matchFrame.pageNumber);

}



// Given an old frame that represents that same page, incorporate
// its old information into this one.  Note that we keep extra
// information about pages to perform a number of verifications on
// the trace being read.
void
PageDumpFrame::update (PageDumpFrame& oldFrame) {

  // Make sure that the old frame does represent the same page as this
  // one.
  assert(match(oldFrame));

  // The new frame's fetched image should match the state of the page
  // as it was evicted from the unprotected set last time.  So, if the
  // old frame was clean, the new frame's fetched image should match
  // the old frame's fetched image, and if the old frame was dirty,
  // the new frame's fetched image should match the old frame's
  // evicted image.  If they don't match, something is wrong with the
  // tracer.
  int comparisonResult;
  if (oldFrame.dirty) {

    comparisonResult = memcmp(fetchedImage,
			      oldFrame.evictedImage,
			      BYTES_PER_PAGE);

    // If the new frame represents a dirtied page, then the new
    // frame's evict dump image is the one that needs to be kept, as
    // it represents what would be evicted if this page were extracted
    // from the LRU queue.  If the new frame does not represent a
    // dirtied page, then old frame's evict dump image should be kept,
    // as that is what would be evicted from the LRU queue.
    if (!dirty) {
      memcpy(evictedImage, oldFrame.evictedImage, BYTES_PER_PAGE);
    }

    // If the old frame was dirty, than then this new one should also be
    // dirty.  If the old frame was clean, then the new frame should
    // retain whatever dirtiness setting it obtained from the record
    // that was read.
    dirty = true;

  } else {

    // The old frame was clean, which means that during the time that
    // this page would spend in an LRU-managed memory, it would still
    // be clean.  So, we need perform no updates, but merely check
    // that the last recorded image of this page matched this one.
    comparisonResult = memcmp(fetchedImage,
			      oldFrame.fetchedImage,
			      BYTES_PER_PAGE);

  }

  if (comparisonResult != 0) {
    fprintf(stderr,
	    "New fetch doesn't match appropriate old image!\n");
    exit(-1);
  }
}



// Output this frame's fetch information.  Merely output the page
// number and the fetched image of this page.  This information will
// reflect the traffic that would be required to fetch this page into
// an LRU-managed memory.
void
PageDumpFrame::outputFetch (FILE* outputStream) {

  char compulsoryField = (compulsory ? 'C' : 'F');
  fprintf(outputStream, "%1c%4x", compulsoryField, pageNumber);

  // Output the fetch dump image.
  safe_fwrite((void*)fetchedImage,
	      sizeof(word),
	      WORDS_PER_PAGE,
	      outputStream);
}



// Output this frame's eviction information.  Output the page number,
// dirty field, and evicted image of this page.  This information will
// reflect the traffic that would be required to evict this page from
// an LRU-managed memory.  Note that the image is output irrespective
// of dirtiness, as the next level in the memory heirarchy may not
// keep copies of pages, so even clean pages will need to be
// transferred.
void
PageDumpFrame::outputEvict (FILE* outputStream) {

  // Output the dirtiness and page number information.
  char dirtyField = (dirty ? 'W' : 'R');
  fprintf(outputStream, "%1c%4x", dirtyField, pageNumber);

  // If this page has been dirtied, we need to output the evicted
  // image, which will be different from the fetched image.  However,
  // if the page is still clean, we output the fetched image, as that
  // would be what a paging system would evict given a clean page.
  void* outputImage;
  if (dirty) {

    outputImage = (void*)evictedImage;

  } else {

    outputImage = (void*)fetchedImage;

  }

  safe_fwrite((void*)outputImage,
	      sizeof(word),
	      WORDS_PER_PAGE,
	      outputStream);
}
