// filter.cc
// Scott F. Kaplan -- sfkaplan@cs.utexas.edu
// August 1997

// Accept the output of from VMTrace with the page-dump experiment and
// reduce it based on the LSD trace filter.  Thus, each outgoing
// record will contain information and page images for each page
// fetched into and evicted from a memory of the given size.



#include <stdio.h>
#include <list.h>
#include <set.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <assert.h>
#include "PageDumpFrame.hh"



list<PageDumpFrame>::iterator
searchQueue (list<PageDumpFrame>& LRUQueue, PageDumpFrame& matchFrame) {

  // Loop through the LRU queue, searching for a frame that matches
  // the given frame.  ("Matching" is defined by the PageDumpFrame
  // class.)
  list<PageDumpFrame>::iterator currentIterator = LRUQueue.begin();
  while (currentIterator != LRUQueue.end()) {
    if (matchFrame.match(*currentIterator)) {

      // Match found.  Return this iterator.
      return currentIterator;

    } else {

      // Not a match.  Move to the next item in the LRU queue.
      currentIterator++;

    }
  }

  return currentIterator;

}


// Given a VMTrace generated trace that contains images of each page
// as it was fetched into and evicted from the unprotected set, read
// each record and filter it through an LRU queue.  The output
// generated reflects the traffic into and out of an LRU-managed
// memory of the given size.
void
processTrace (FILE* inputStream,
	      FILE* outputStream,
	      int numberOfPages) {
  list<PageDumpFrame> LRUQueue;
  set<unsigned int, less<unsigned int> > knownPages;
  // A little debugging code.
  unsigned int compulsoryFaults = 0;
  unsigned int regularFaults = 0;
  unsigned int writeBacks = 0;

  // Read through the trace until there is nothing left in it.  At
  // each record, read that record and then apply it to the LRU
  // queue.
  bool done = false;
  do {

    // Read a character and then put it back into the input stream.
    // We do this in order to set the end-of-file flag in the stream
    // object.  Note: Since it seems that the eof flag isn't set
    // properly, at least under Linux, we perform a check for an
    // explicit end-of-trace character, 'Q'.  It's an ugly workaround,
    // but it is reliable.
    char testChar = (char)fgetc(inputStream);
    ungetc((int)testChar, inputStream);
    if (testChar == 'Q') {

      done = true;

      // The end of the trace has been reached, so write the
      // end-of-trace tag to the output.
      fputc((int)'Q', outputStream);

    } else {

      // Read the next record into a dump frame.
      PageDumpFrame readFrame;
      readFrame.load(inputStream);
      
      // Does this record correspond to a page that is already in the
      // LRU queue?  Do whatever specific work is required for these
      // two cases.
      list<PageDumpFrame>::iterator searchIterator;
      searchIterator = searchQueue(LRUQueue, readFrame);
      if (searchIterator != LRUQueue.end()) {
	
	// The page is already in the LRU queue, so no traffic into or
	// out of an LRU-managed memory would be generated.  Update the
	// new frame based on the old frame for this page, and then
	// erase the old page.
	readFrame.update(*searchIterator);
	LRUQueue.erase(searchIterator);
	
      } else {
	
	// The page is not in the LRU queue, so this record represents a
	// fetch into an LRU-managed memory of this size.  Output the
	// fetched image of the newly read page to represent incoming
	// traffic.  (Note that even compulsory faults are output,
	// although with a tag to indicate that it is compulsory.  With
	// that information, a simulator can ignore such fetch
	// information if it wants to.)  Also, if the LRU queue is full
	// (the list is as big as the memory to be simulated), we must
	// select a page for eviction.  Output the evicted image and
	// dirty field of the page selected for eviction to represent
	// outgoing traffic.
	assert (LRUQueue.size() <= numberOfPages);
	if (knownPages.find(readFrame.getPageNumber()) !=
	    knownPages.end()) {
	  // This page has been seen before, so this fetch information
	  // represents a genuine fault.
	  readFrame.setCompulsory(false);
	  regularFaults++;
	} else {
	  // This page has not been seen before, so this is a compulsory
	  // fault.
	  knownPages.insert(readFrame.getPageNumber());
	  readFrame.setCompulsory(true);
	  compulsoryFaults++;
	}
	readFrame.outputFetch(outputStream);
	if (LRUQueue.size() == numberOfPages) {
	  list<PageDumpFrame>::iterator lastIterator = LRUQueue.end();
	  lastIterator--;
	  (*lastIterator).outputEvict(outputStream);
	  // A little debugging code.
	  if ((*lastIterator).isDirty()) {
	    writeBacks++;
	  }
	  LRUQueue.erase(lastIterator);
	} else {
	  // Since there will be no evicted page image, output a tag to
	  // provide that information.
	  fprintf(outputStream, "N0000");
	}
      }
      
      // Insert the new frame for this page at the head of the LRU
      // queue.
      LRUQueue.push_front(readFrame);

    }

  } while (!done);

  // Since pages that remain in memory at the end of execution
  // disappear, we don't need to flush the LRU queue here.  (If we're
  // simulating something like a persistent store, then we would.)

  // A little debugging code.
  printf("Compulsory faults: %d\n", compulsoryFaults);
  printf("Regular faults:    %d\n", regularFaults);
  printf("Write backs:       %d\n", writeBacks);

}



void
main (int argc, char** argv) {

  if (argc != 4) {
    cerr << "Usage: "
	 << argv[0]
	 << " <input filename>"
	 << " <output filename>"
	 << " <memory size in pages>"
	 << endl;
    exit(-1);
  }

  // Open the input and output files.
  FILE* inputStream = fopen(argv[1], "r+");
  if (inputStream == 0) {
    perror("Could not open input file");
  }
  FILE* outputStream = fopen(argv[2], "w+");
  if (outputStream == 0) {
    perror("Could not open output file");
  }

  // Extract the desired memory size.
  int numberOfPages = atoi(argv[3]);
  if (numberOfPages <= 0) {
    cerr << "Invalid memory size " << argv[3] << endl;
    exit(-1);
  }

  // Initiate the trace processing loop.
  processTrace (inputStream, outputStream, numberOfPages);

  fclose(inputStream);
  fclose(outputStream);

}
